#include <stdio.h>  
#include <stdlib.h>
#include <string.h> 
#include <unistd.h> 
#include <fcntl.h>  
#include <errno.h>  
#include <termios.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/param.h>
#include <time.h>
#include <math.h>
#include <inttypes.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/wait.h>
#include <signal.h>

#define RADIO_DEV "/dev/ttyUSB0"
#define BUFFER_MAX 200
#define RECORD_NAME "/var/run/nwx.dat"
#define LOG_NAME "/data/nwx/nwx.log"

#define LISTEN_SOCKET  "3000"
#define SOCKET_BACKLOG 1

int open_radio(void);
void read_data(void);
void receive_response(void);
void send_request(unsigned char *);
void background(void);
void log_msg(unsigned char *,unsigned char *);
int fix_wind_dir(int);
char *dir_to_name(int);
void write_data(void);
void create_peetbros(void);
void convert_value( int16_t, char * );
void convert_value_short( int16_t, char * );
void dump_peetbros(void);
int open_socket(void);
void check_socket(void);
void send_peetbros(void);
void sigpipe_handler(int);
void *get_in_addr(struct sockaddr *);
  
int serial_fd=-1;
int bufptr = 0;
unsigned char buffer[BUFFER_MAX];
unsigned char errbuf[BUFFER_MAX];
unsigned char peet_bros[BUFFER_MAX];
int background_mode=0;

double pressure, temp, humidity, wind_speed, wind_array[4], wind_average;
int raw_wind, raw_rain, raw_wind_vane, cycles, wind_dir;

int sockfd=-1, client_fd=-1; 
struct sigaction sa;

int main(int argc, char *argv[])
{
  int i;

  background_mode=1; /* default is to go into background */
  memset(buffer,0,BUFFER_MAX);

  for(i=0;i<argc;i++)
    {
      if(!strcmp(argv[i],"-v"))
	{
	  background_mode=0;
	}
    }

  if(background_mode)
    background();

  if( !open_radio() )
    exit(-1);

  if( !open_socket() )
    exit(-1);
  
  /* for computing 1 minute average, used in Peet Bros format */
  wind_array[0]=0.0;
  wind_array[1]=0.0;
  wind_array[2]=0.0;
  wind_array[4]=0.0;
  wind_average=0.0;
  
  log_msg("CTL ","NANO-WX driver starting.");

  while(1)
    {
      read_data();
      check_socket();
    }
}


/*
 * Open serial port 433Mhz radio is connected to, set to 2400 8N1, and
 * disable all HW and SW flow control.
 */
int open_radio(void)
{
  struct termios options;

  serial_fd = open(RADIO_DEV, O_RDWR | O_NOCTTY | O_NDELAY);
  if (serial_fd == -1)
    {
      perror("Unable to open " RADIO_DEV );
      return 0;
    }

  /* set non-blocking mode */
  fcntl(serial_fd, F_SETFL, FNDELAY);

  /* Get the current options for the port */
  tcgetattr(serial_fd, &options);

  /* Set the baud rate to 9600 */
  cfsetispeed(&options, B2400);
  cfsetospeed(&options, B2400);

  /* Enable the receiver and set local mode  */
  options.c_cflag |= (CLOCAL | CREAD);

  /* Set 8N1 */
  options.c_cflag &= ~PARENB;
  options.c_cflag &= ~CSTOPB;
  options.c_cflag &= ~CSIZE;
  options.c_cflag |= CS8;

  /* Set no HW flow control */
  /*options.c_cflag &= ~CNEW_RTSCTS;*/

  /* Set no SW flow control */
  options.c_iflag &= ~(IXON | IXOFF | IXANY);

  /* Set raw input */
  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG);

  /* Set raw output */
  options.c_oflag &= ~OPOST;

  /* Set the new options for the port */
  tcsetattr(serial_fd, TCSAFLUSH, &options);

  return 1;
}


/*
 * Every 5 seconds, read the data values from the NANO WX
 * Every 15 seconds, reset the wind values
 * At the top of every minute, reset the rain values
 */
void read_data(void)
{
  time_t now;
  struct tm *time_now;
  int i, seconds;
  unsigned char ch;

  time(&now);
  time_now=localtime(&now);
  seconds = time_now->tm_sec;
  
  /* flush any data that has come in */
  while( read(serial_fd,&ch,1) > 0 );
  
  /* every 5 seconds */
  if( (time_now->tm_sec % 5) == 0 )
    {
      /* send data request */
      send_request("D");
      receive_response();
      write_data();
    }

  /* every 15 seconds */
  if( (time_now->tm_sec % 15) == 0 )
    {
      send_request("W"); /* zero wind counter */
      receive_response();
      wind_speed = (raw_wind * 1.492) / 15.0;

      /* compute one minute wind average - used by Peet Bros format */
      wind_array[0] = wind_array[1];
      wind_array[1] = wind_array[2];
      wind_array[2] = wind_array[3];
      wind_array[3] = wind_speed;
      wind_average = ( wind_array[0] + wind_array[1] + wind_array[2] + wind_array[3] ) / 4.0;
      
      create_peetbros();
      /*dump_peetbros();*/
      log_msg("PEET ",peet_bros);

      send_peetbros(); /* send peet bros data to network (if connected) */
      
      write_data(); /* write data to file */
    }

  /* every day, at midnight, reset rain total */
  if( (time_now->tm_hour == 0) && (time_now->tm_min == 0) && (time_now->tm_sec == 0 ) )
    {
      send_request("R"); /* zero rain counter */
      receive_response();
      write_data();
    }

  /* wait until next second to avoid double triggering */
  while( seconds == time_now->tm_sec )
    {
      time(&now);
      time_now=localtime(&now);
    }
}

/*
 * Receive data from the serial port. Time out after 1 second
 */
void receive_response(void)
{
  unsigned char ch;
  fd_set readfds;
  struct timeval timeout;
  char *tok;
  
  memset(buffer,0,BUFFER_MAX);
  bufptr=0;
  ch = 0;
  while( ch != 0x0a ) /* strings from NANO-WX end in 0x0d 0x0a */
    {
      FD_CLR(serial_fd,&readfds);
      FD_SET(serial_fd,&readfds);
      timeout.tv_sec=1;
      timeout.tv_usec=100000;
      select(serial_fd+1,&readfds,NULL,NULL,&timeout);
      if( FD_ISSET( serial_fd, &readfds ) )
	{
	  if( read(serial_fd,&ch,1) > 0 )
	    {
	      /*printf("|%c [%x]|",ch,ch);*/
	      if( ch != '\n' && ch != '\r' )
		{
		  if( bufptr < BUFFER_MAX )
		    buffer[bufptr++]=ch;
		}
	    }
	}
      else
	{
	  /* we hit a timeout without a terminator on the string .. log the problem */
	  snprintf(errbuf, BUFFER_MAX, "RX incomplete string '%s'",buffer);
	  log_msg("ERR  ",errbuf);
	  return;
	}
    }

  /* log the string we got */
  log_msg("RX < ",buffer);
  
  /* if we made it to here, we have recieved a terminated string back from the NANO-WX */
  /* process it */
  if( strncasecmp(buffer,"NANO-WX DATA",12) == 0 )
    {
      /* the response contains weather data - parse it */
      // NANO-WX DATA 976.49 43.38 64.99 54 0 875 25
      if( (tok=strtok(buffer+13," ")) != NULL )
	{
	  pressure=strtof(tok,NULL);
	  if( (tok=strtok(NULL," ")) != NULL )
	    {
	      humidity=strtof(tok,NULL);
	      if( (tok=strtok(NULL," ")) != NULL )
		{
		  temp=strtof(tok,NULL);
		  if( (tok=strtok(NULL," ")) != NULL )
		    {
		      raw_wind = strtol(tok,NULL,10);
		      if( (tok=strtok(NULL," ")) != NULL )
			{
			  raw_rain = strtol(tok,NULL,10);
			  if( (tok=strtok(NULL," ")) != NULL )
			    {
			      raw_wind_vane = strtol(tok,NULL,10);
			      if( (tok=strtok(NULL," ")) != NULL )
				{
				  cycles = strtol(tok,NULL,10);
				}
			    }
			}
		    }
		}
	    }
	}

      wind_dir = fix_wind_dir(raw_wind_vane);
      
#ifdef DEBUG_PRINT_VALUES
      printf("Pressure = %6.2f mb\n",pressure);
      printf("Temp     = %5.2f F\n",temp);
      printf("Humidity = %4.2f %c\n",humidity,'%');
      printf("Wind     = %5.2f MPH\n", (raw_wind * 1.492) / (float)cycles );
      printf("Wind Dir = %s [%d deg]\n",dir_to_name(wind_dir), wind_dir);
      printf("Rain     = %5.2f in\n",raw_rain * 0.11);
#endif
      
    } /* end if DATA record */
}


/*
 * Send message to sensors via radio
 */
void send_request(unsigned char *buf)
{
  write(serial_fd,buf,strlen(buf));
  fsync(serial_fd);
  log_msg("TX > ",buf);
}

/*
 * Log a message to the sensor log file, and the screen if we are not in
 * background mode.
 */
void log_msg(unsigned char *prefix, unsigned char *ptr)
{
  time_t now;
  struct tm *time_now;
  FILE *log_file=NULL;

  if(!prefix || !ptr)
    return;

  time(&now);
  time_now=localtime(&now);

  if(!background_mode)
    {
      printf("%02d/%02d/%02d %02d:%02d:%02d | %s%s\n",
	     time_now->tm_mon+1,
	     time_now->tm_mday,
	     time_now->tm_year+1900,
	     time_now->tm_hour,
	     time_now->tm_min,
	     time_now->tm_sec,
	     prefix,
	     ptr);
      fflush(stdout);
    }

  if((log_file=fopen(LOG_NAME,"a+"))!=NULL)
    {
      fprintf(log_file,"%02d/%02d/%02d %02d:%02d:%02d | %s%s\n",
	      time_now->tm_mon+1,
	      time_now->tm_mday,
	      time_now->tm_year+1900,
	      time_now->tm_hour,
	      time_now->tm_min,
	      time_now->tm_sec,
	      prefix,
	      ptr);
      fflush(log_file);
      fclose(log_file);
    }
}


/*
 * Move this process into the background.
 */
void background(void)
{
  int childpid, fd;
  
  /*
   * If we were started by init (process 1) from the /etc/inittab file
   * there's no need to detach.
   * This test is unreliable due to an unavoidable ambiguity
   * if the process is started by some other process and orphaned
   * (i.e., if the parent process terminates before we are started).
   */
  
  if (getppid() != 1)
    {
      /*
       * If we were not started in the background, fork and
       * let the parent exit.  This also guarantees the first child
       * is not a process group leader.
       */
      
      if ( (childpid = fork()) < 0)
	{
	  printf("Can't fork first child\n");
	}
      else if (childpid > 0)
	{
	  exit(0);/* parent */
	}
      
      /*
       * First child process.
       *
       * Disassociate from controlling terminal and process group.
       * Ensure the process can't reacquire a new controlling terminal.
       */
      
      if (setpgid(0,0) == -1)
	{
	  printf("Can't change process group\n");
	}
      
      if ( (fd = open("/dev/tty", O_RDWR)) >= 0) 
	{
	  ioctl(fd, TIOCNOTTY, (char *)NULL); /* lose controlling tty */
	  close(fd);
	}
      
    } /* end if( getppid() != 1 ) */
  
  /*
   * Close any open file descriptors.
   */
  
  for (fd = 0; fd < NOFILE; fd++)
    {
      close(fd);
    }
  
  errno = 0;/* probably got set to EBADF from a close */
}

static int previous_heading;

int fix_wind_dir(int raw_wind_dir)
{
  /*  Data from raw readings:
   *  N  786		
   *  NE   460
   *  E   93
   *  SE  184
   *  S   287
   *  SW  630
   *  W   940 938-944
   *  NW  886
   */
  
  int dir_low[] =    { 760, 440, 70, 160, 260, 610, 920, 860 };
  int dir_center[] = { 786, 460, 93, 184, 287, 630, 940, 886 };
  int dir_hi[] =     { 805, 480, 115,205, 305, 650, 960, 900 };
  int headings[]  =  {   0,  45,  90, 135,180, 225, 270, 315 };
  int index,min_diff,min_diff_index,diff;

  /* try to find a low/hi interval the reading fits within */
  for(index=0;index<=7;index++)
    {
      if( (dir_low[index] <= raw_wind_dir) && (raw_wind_dir <= dir_hi[index]) )
	{
	  previous_heading = headings[index];
	  return headings[index];
	}
    }

#ifdef USE_MIN_DIFF
  /* if we made it to here, the reading did not fit in the intervals. find */
  /* the minimum difference from the center values, and use that */
  min_diff=359;
  for(index=0;index<=7;index++)
    {
      diff=abs(dir_center[index]-raw_wind_dir);
      if( diff < min_diff )
	{
	  min_diff = diff;
	  min_diff_index = index;
	}
    }
  return headings[min_diff_index];
#else
  return previous_heading;
#endif
}

char *dir_to_name(int heading)
{
  switch(heading)
    {
    case 0:
      return "N";
      break;
    case 45:
      return "NE";
      break;
    case 90:
      return "E";
      break;
    case 135:
      return "SE";
      break;
    case 180:
      return "S";
      break;
    case 225:
      return "SW";
      break;
    case 270:
      return "W";
      break;
    case 315:
      return "NW";
      break;
    }
}

void write_data(void)
{
  time_t now;
  FILE *data_file = NULL;

  time(&now);
  
  if((data_file=fopen(RECORD_NAME,"w"))!=NULL)
    {
      fprintf(data_file,"%6.2f\n",pressure);
      fprintf(data_file,"%5.2f\n",temp);
      fprintf(data_file,"%4.2f\n",humidity);
      fprintf(data_file,"%5.2f\n",wind_speed);
      fprintf(data_file,"%d\n",wind_dir);
      fprintf(data_file,"%5.2f\n",raw_rain * 0.011);
      fprintf(data_file,"%5.2f\n",wind_average);
      fprintf(data_file,"%d\n",now);
      fflush(data_file);
      fclose(data_file);
    }
}

/* nwx-driver emulates a ADS-WS1 emittng peet bros format data */
/* on TCP port WXPORT. weewx can connect to that port and read */
/* peet bros format weather data, just like it was talking to  */
/* a WS1. */

/*      WS1 station emits data in PeetBros format:
 *
 *      http://www.peetbros.com/shop/custom.aspx?recid=29
 *
 *      Each line has 50 characters - 2 header bytes and 48 data bytes:
 *
 *      !!000000BE02EB000027700000023A023A0025005800000000
 *        SSSSXXDDTTTTLLLLPPPPttttHHHHhhhhddddmmmmRRRRWWWW
 *
 *        SSSS - wind speed (0.1 kph)
 *        XX   - wind direction calibration
 *        DD   - wind direction (0-255)
 *        TTTT - outdoor temperature (0.1 F)
 *        LLLL - long term rain (0.01 in)
 *        PPPP - pressure (0.1 mbar)
 *        tttt - indoor temperature (0.1 F)
 *        HHHH - outdoor humidity (0.1 %)
 *        hhhh - indoor humidity (0.1 %)
 *        dddd - date (day of year)
 *        mmmm - time (minute of day)
 *        RRRR - daily rain (0.01 in)
 *        WWWW - one minute wind average (0.1 kph)
 *
 * Notes from Peet Bros website:
 * 1. Unless otherwise specified, the units of weather data are always as 
 * follows:  Wind Speed 0.1kph; Wind Direction 0-255 (corresponding to 0-360
 * degrees compass heading, North = 0), Temperature 0.1 degrees F (also Heat
 * Index, Wind Chill, and Dew Point), Rain 0.01 inches, Humidity 0.1%, 
 * Barometer 0.1mbar.
 * 2. The first two digits of Wind Direction fields are normally "00" but if
 * a wind direction calibration number has been entered, they may be "FF".
 * 3. Negative values are output in "twos complement" format.
 * 3. A Time value of 0000 is midnight; a Date value of 0000 is January 1.
 * 4. The Leap Year Counter (pre-2004 models) determines whether the year has
 * 365 or 366 days.  On these models, there is no "year" stored in memory 
 * (only month and day).
 * 5. Rain Gauges that measure in increments of 0.1mm are fully supported on 
 * the ULTIMETER keyboard display, but please be advised that when 0.1mm 
 * increment is selected, the serial data rain output is in 0.1mm increments.
 * Serial data rain output for all other rain gauge selections is reported in increments of 0.01 inches. 
 */
void create_peetbros(void)
{
  double tmp;
  int16_t value;
  int32_t value2;
  time_t now;
  struct tm *time_now;

  memset(peet_bros,0,BUFFER_MAX);

  /* convert wind speed - converts to 'windSpeed' */
  tmp = wind_speed * 1.609344; /* convert to kph */
  tmp *= 10; /* shift it to 0.1kph units */
  value = tmp;
  convert_value(value,peet_bros);

  /* wind direction calibration - ignored by weewx */
  peet_bros[4]='0';
  peet_bros[5]='0';

  /* convert wind direction - converts to 'windDir' */
  value2 = (wind_dir * 255) / 360;
  convert_value_short( (int16_t)value2, peet_bros+6 );

  /* convert outdoor temp - converts to 'outTemp' */
  tmp = temp * 10; /* convert to 0.1F units */
  value = tmp;
  convert_value( value, peet_bros+8 );

  /* convert rain .. this ends up as 'rain_total' */
  tmp = raw_rain * 0.011 * 100; /* conver to 0.01 in increments */
  value = tmp;
  convert_value( value, peet_bros+12 );

  /* convert pressure - ends up as 'pressure' */
  tmp = pressure * 10; /* convert to 0.1mb */
  value = tmp;
  convert_value( value, peet_bros+16 );

  /* indoor temp = outdoor temp - coverts to 'inTemp' */
  tmp = temp * 10; /* convert to 0.1F units */
  value = tmp;
  convert_value( value, peet_bros+20 );

  /* outdoor humidity + indoor hiumidity */
  tmp = humidity * 10; /* convert to 0.1% */
  value = tmp;
  convert_value( value, peet_bros+24 ); /* 'outHumidity' */
  convert_value( value, peet_bros+28 ); /* 'inHumidity' */

  /* day of year + minute of day */
  time(&now);
  time_now=localtime(&now);  
  value = time_now->tm_yday;
  convert_value( value, peet_bros+32 ); /* 'day_of_year' */
  value = time_now->tm_min;
  convert_value( value, peet_bros+36 ); /* 'minute_of_day' */

  /* convert rain .. this ends up as 'daily_rain' */
  tmp = raw_rain * 0.011 * 100; /* conver to 0.01 in increments */
  value = tmp;
  convert_value( value, peet_bros+40 );

  /* convert 1 minute wind average - 'wind_average' */
  tmp = wind_average * 1.609344; /* convert to kph */
  tmp *= 10; /* shift it to 0.1kph units */
  value = tmp;
  convert_value(value,peet_bros+44);  
}

/* take an integer value, convert to a 2 digit hex value */
void convert_value_short( int16_t value, char *buff )
{
  uint16_t temp;

  temp = value;
  //if( value < 0 )
  //  {
  //    temp = (~value) + 1;
  //  }
  //printf("convert short in=%d  out=%d\n",value,temp);
  snprintf(buff,3,"%02X",temp);
}

/* take an integer value, convert to a 4 digit hex value */
void convert_value( int16_t value, char *buff )
{
  uint16_t temp;

  temp = value;
  //if( value < 0 )
  //#  {
  //#    temp = (~value) + 1;
  //#  }
  //printf("convert in=%d  out=%d\n",value,temp);
  snprintf(buff,5,"%04X",temp);
}

void dump_peetbros(void)
{
  int i;
  for(i=0;i<48;i++)
    {
      if( peet_bros[i] >= 32 )
	printf("%c",peet_bros[i]);
      else
	printf(".");
    }
  printf("    ");
  for(i=0;i<48;i++)
    {
      printf("%2x ",peet_bros[i]);
    }
  printf("\n");
}

/* --------------------- SOCKET CODE -------------------------- */

int open_socket(void)
{
  struct addrinfo hints, *servinfo, *p;
  int rv;
  int yes=1;
  
  memset(&hints, 0, sizeof hints);
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_PASSIVE; // use my IP

  if ((rv = getaddrinfo(NULL, LISTEN_SOCKET, &hints, &servinfo)) != 0) {
    fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(rv));
    return 0;
  }

  // loop through all the results and bind to the first we can
  for(p = servinfo; p != NULL; p = p->ai_next) {
    if ((sockfd = socket(p->ai_family, p->ai_socktype,
			 p->ai_protocol)) == -1) {
      perror("server: socket");
      continue;
    }

    if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes,
		   sizeof(int)) == -1) {
      perror("setsockopt");
      return 0;
    }

    if (bind(sockfd, p->ai_addr, p->ai_addrlen) == -1) {
      close(sockfd);
      perror("server: bind");
      continue;
    }

    break;
  }

  freeaddrinfo(servinfo); // all done with this structure

  if (p == NULL)  {
    fprintf(stderr, "server: failed to bind\n");
    return 0;
  }

  if (listen(sockfd, SOCKET_BACKLOG) == -1) {
    perror("listen");
    return 0;
  }

  sa.sa_handler = sigpipe_handler; /* handle closed sockets */
  sigemptyset(&sa.sa_mask);
  sa.sa_flags = SA_RESTART;
  if (sigaction(SIGPIPE, &sa, NULL) == -1) {
    perror("sigaction");
    return 0;
  }

  return 1;
}

/* check to see if client connection waiting - if so, accept */
void check_socket(void)
{
  fd_set readfds;
  struct timeval timeout;
  struct sockaddr_storage their_addr;
  socklen_t sin_size;
  char s[INET6_ADDRSTRLEN];

  if( client_fd != -1 )  /* already an active client connection */
    return;
  
  FD_CLR(sockfd,&readfds);
  FD_SET(sockfd,&readfds);
  timeout.tv_sec=0;
  timeout.tv_usec=100000;
  select(sockfd+1,&readfds,NULL,NULL,&timeout);
  if( FD_ISSET( sockfd, &readfds ) )
    {
      sin_size = sizeof their_addr;
      client_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size);
      if (client_fd == -1) {
	return;
      }

    inet_ntop(their_addr.ss_family,
	      get_in_addr((struct sockaddr *)&their_addr),
	      s, sizeof s);
    snprintf(buffer,BUFFER_MAX-1,"Connection from %s", s);
    log_msg("INFO ",buffer);
    }
}

void send_peetbros(void)
{
  if( client_fd != -1 )
    {
      if( send(client_fd,"!!",2,0) == -1 )
	{
	  close(client_fd);
	  client_fd=-1;
	  log_msg("INFO ","Unable to write !! - closing client connection");
	}
      else
	{
	  if( send(client_fd,peet_bros,48,0) == -1 )
	    {
	      close(client_fd);
	      client_fd=-1;
	      log_msg("INFO ","Unable to write PEET - closing client connection");
	    }
	  else
	    {
	      if( send(client_fd,"\r\n",2,0) == -1 )
		{
		  close(client_fd);
		  client_fd=-1;
		  log_msg("INFO ","Unable to write CRLF - closing client connection");
		}
	    }
	}
    }
}

void sigpipe_handler(int s)
{
  /* if we've gotten a SIGPIPE, it is most likely from the client closing */
  /* the TCP connection */

  /* close the file handle, and go back to listening for connections */
  close(client_fd);
  client_fd=-1;
}

// get sockaddr, IPv4 or IPv6:
void *get_in_addr(struct sockaddr *sa)
{
  if (sa->sa_family == AF_INET) {
    return &(((struct sockaddr_in*)sa)->sin_addr);
  }

  return &(((struct sockaddr_in6*)sa)->sin6_addr);
}
